from __future__ import unicode_literals

from django.contrib.contenttypes.models import ContentType
from django.core import serializers
from django.core.paginator import Paginator, EmptyPage, PageNotAnInteger
from django.db import models

import json


class SerializableMixin(models.Model):
    def to_json(self):
        '''
        Default dictionary representation of a model that can later be converted to JSON.
        May be overriden in a model by defining `to_json` in the model.
        '''
        serialized_model_str = serializers.serialize('json', [self, ], cls=serializers.json.DjangoJSONEncoder)
        serialized_model = json.loads(serialized_model_str)[0]

        pk = serialized_model.get('pk')
        serialized_model = serialized_model.get('fields')
        serialized_model['pk'] = pk

        return serialized_model

    @staticmethod
    def serialize(queryset_or_model, page_number=1, limit=None):
        '''
        Serializes a queryset/model into JSONifable dictionary.
        TODO: Add this to queryset methods instead of being a static method.

        Example: ```{
            'user': {
                'models': [{'pk': 1, 'email': '', }, ],
                'count': 1,
                'pageNumber': 1,
                'totalCount': 1,
                'perPage': 10,
                'numberOfPages': 1,
                'firstRecordNumber': 1,
                'lastRecordNumber': 1,
            }
        }
        ```
        }
        '''
        queryset = None
        model = None
        queryset_page = None
        count = 0

        if isinstance(queryset_or_model, models.Model):
            # handle just one model being passed in as a queryset
            model = queryset_or_model
            queryset = [model]
        elif isinstance(queryset_or_model, models.query.QuerySet):
            queryset = queryset_or_model
            model = queryset.model

            if limit > 1:
                paginator = Paginator(queryset, limit)

                try:
                    queryset_page = paginator.page(page_number)
                except PageNotAnInteger:
                    queryset_page = paginator.page(1)
                except EmptyPage:
                    queryset_page = paginator.page(paginator.num_pages)

                queryset = queryset_page.object_list
        else:
            raise NotImplementedError()

        content_type = ContentType.objects.get_for_model(model)
        model_name = content_type.name
        serialized_models = [m.to_json() for m in queryset]
        count = len(serialized_models)

        serialized_value = {
            model_name: {
                'models': serialized_models,
                'count': count,
                'pageNumber': 1,
                'totalCount': count,
                'perPage': count,
                'numberOfPages': 1,
                'firstRecordNumber': 0,
                'lastRecordNumber': count,
            }
        }

        if count == 1:
            serialized_value[model_name]['model'] = serialized_models[0]

        if queryset_page:
            first_record_number = (queryset_page.number - 1) * queryset_page.paginator.per_page + 1

            if queryset_page.paginator.count == 0:
                first_record_number = 0

            last_record_number = queryset_page.number * queryset_page.paginator.per_page

            if last_record_number > queryset_page.paginator.count:
                last_record_number = queryset_page.paginator.count

            serialized_value[model_name].update({
                'pageNumber': queryset_page.number,
                'totalCount': queryset_page.paginator.count,
                'perPage': queryset_page.paginator.per_page,
                'numberOfPages': queryset_page.paginator.num_pages,
                'firstRecordNumber': first_record_number,
                'lastRecordNumber': last_record_number,
            })

        return serialized_value

    class Meta:
        app_label = 'djajax'
        abstract = True



from __future__ import unicode_literals

from django.db import models
from django.test import SimpleTestCase, TestCase

from ..models import SerializableMixin


class FakeModel(SerializableMixin):
    name = models.CharField(max_length=10)


class SerializableMixinTestCase(TestCase):
    def setUp(self):
        pass

    def test_model_serialize_model(self):
        model = FakeModel(pk=1, name='fake')

        expected = {
            u'fake model': {
                u'count': 1,
                u'firstRecordNumber': 0,
                u'lastRecordNumber': 1,
                u'model': {
                    u'name': u'fake', u'pk': 1
                },
                u'models': [{
                    u'name': u'fake', u'pk': 1
                }, ],
                u'numberOfPages': 1,
                u'pageNumber': 1,
                u'perPage': 1,
                u'totalCount': 1
            }
        }
        self.assertEqual(expected, model.serialize(model))

    def test_model_serialize_queryset(self):
        self.maxDiff = None
        model_one = FakeModel(name='fake 1')
        model_one.save()

        model_two = FakeModel(name='fake 2')
        model_two.save()

        all_models = FakeModel.objects.filter()

        expected = {
            u'fake model': {
                u'count': 2,
                u'firstRecordNumber': 0,
                u'lastRecordNumber': 2,
                u'models': [{
                    u'name': u'fake 1', u'pk': model_one.pk
                }, {
                    u'name': u'fake 2', u'pk': model_two.pk
                }, ],
                u'numberOfPages': 1,
                u'pageNumber': 1,
                u'perPage': 2,
                u'totalCount': 2
            }
        }
        self.assertEqual(expected, FakeModel.serialize(all_models))


class SerializableMixinSimpleTestCase(SimpleTestCase):
    def setUp(self):
        pass

    def test_model_to_json(self):
        model = FakeModel(pk=1, name='fake')
        expected = {
            u'name': u'fake',
            u'pk': 1
        }
        self.assertEqual(expected, model.to_json())



from django.contrib.contenttypes.models import ContentType

from .http import JsonHttpResponse

import json


def is_idempotent(request):
    if request.method == 'POST' and request.body:
        json_body = {}

        try:
            json_body = json.loads(request.body)
        except:
            pass

        if json_body:
            create_json = json_body.get('create')
            update_json = json_body.get('update')
            delete_json = json_body.get('delete')

            if create_json or update_json or delete_json:
                return False

    return True


def response(request, extra={}):
    if request.is_ajax():
        if request.method == 'GET':
            return get(request, extra=extra)
        elif request.method == 'POST':
            return post(request, extra=extra)
        else:
            raise NotImplementedError('Only GET and POST are supported')


def get(request, extra={}):
    res = {}

    if extra:
        res.update(extra)

    return JsonHttpResponse(request, res)


def _explode_fields(model, model_json):
    '''
    Handles traversing relationships if there are `__` in the JSON key
    '''
    json_dict = {}

    for field in model_json.keys():
        if '__' in field:
            val = model_json[field]
            related_model_field_name = field.split('__')[0]
            RelatedModel = model._meta.get_field(related_model_field_name).related_model

            related_filter = field.replace(related_model_field_name + '__', '')
            related_model = RelatedModel.objects.get(**{related_filter: val})
            json_dict[related_model_field_name] = related_model
        else:
            json_dict[field] = model_json[field]

    return json_dict


def post(request, extra={}):
    res = {}

    if request.body:
        json_body = json.loads(request.body)

        if json_body:
            model_json = json_body.get('model')
            model = _get_model_from_json(model_json)

            create_json = json_body.get('create')
            update_json = json_body.get('update')
            delete_json = json_body.get('delete')

            if create_json:
                create_dict = _explode_fields(model, create_json)
                obj = model(**create_dict)
                obj.save()
            elif update_json:
                filters = model_json.get('filters')
                obj = model.objects.filter(**filters)

                update_dict = _explode_fields(model, update_json)
                obj.update(**update_dict)
            elif delete_json:
                raise NotImplementedError()

            res = {
                'status': 'ok',
            }

    if extra:
        res.update(extra)

    return JsonHttpResponse(request, res)


def _get_model_from_json(model_json):
    app_label = model_json.get('appLabel')
    model_name = model_json.get('name')
    assert app_label is not None, 'App label could not be found'
    assert model_name is not None, 'Model name could not be found'

    content_type = ContentType.objects.get_by_natural_key(app_label, model_name)
    assert content_type is not None, 'Model content type could not be found'

    model = content_type.model_class()
    return model


def get_from_post_or_get(request, key):
    '''
    Helper to get a key from the POST, and if that fails, to look in the GET.
    '''
    return request.POST.get(key, request.GET.get(key))



from django.http import HttpResponse

import json


def JsonHttpResponse(request, json_text):
    http_accepts = request.META.get('HTTP_ACCEPT', '').split(',')

    if 'application/json' not in http_accepts:
        # TODO: do something here?
        pass

    if isinstance(json_text, dict):
        json_text = json.dumps(json_text)

    response = HttpResponse(json_text, content_type="application/json; charset=utf-8")
    response['Pragma'] = "no cache"
    response['Cache-Control'] = "no-cache, must-revalidate"
    response['content-length'] = len(json_text)

    return response



from __future__ import unicode_literals

from django import template
from django.conf import settings

try:
    from django.urls import reverse
except ImportError:
    # < django 1.10
    from django.core.urlresolvers import reverse

import json

register = template.Library()


@register.filter
def jsonify(val):
    if not val:
        return '{}'

    if not isinstance(val, str):
        val = json.dumps(val)

    return _make_safe(val)


@register.simple_tag(takes_context=True)
def djajax_urls(context):
    # TODO: Cache djajax_urls?
    urls = {
        'requestUrlAndPath': context['request'].get_full_path(),
        'requestPath': context['request'].path,
        'requestMethod': context['request'].method,
    }

    djajax_url_reverse_lookups = []

    if hasattr(settings, 'DJAJAX_URL_REVERSE_LOOKUPS'):
        djajax_url_reverse_lookups = settings.DJAJAX_URL_REVERSE_LOOKUPS

    for reverse_lookup in djajax_url_reverse_lookups:
        url = reverse(reverse_lookup)

        urls.update({
            reverse_lookup: url,
        })

    string_of_urls = json.dumps(urls)

    return _make_safe(string_of_urls)


def _make_safe(val):
    val = template.defaultfilters.safe(val)
    return val



from django.test import TransactionTestCase
try:
    from django.test.runner import DiscoverRunner as BaseRunner
except ImportError:
    # Django < 1.6 fallback
    from django.test.simple import DjangoTestSuiteRunner as BaseRunner

from mock import patch

# https://www.caktusgroup.com/blog/2013/10/02/skipping-test-db-creation/
# Set TEST_RUNNER to this class to use it


class NoDatabaseMixin(object):
    """
    Test runner mixin which skips the DB setup/teardown
    when there are no subclasses of TransactionTestCase to improve the speed
    of running the tests.
    """

    def build_suite(self, *args, **kwargs):
        """
        Check if any of the tests to run subclasses TransactionTestCase.
        """
        suite = super(NoDatabaseMixin, self).build_suite(*args, **kwargs)
        self._needs_db = any([isinstance(test, TransactionTestCase) for test in suite])
        return suite

    def setup_databases(self, *args, **kwargs):
        """
        Skip test creation if not needed. Ensure that touching the DB raises and
        error.
        """
        if self._needs_db:
            return super(NoDatabaseMixin, self).setup_databases(*args, **kwargs)
        if self.verbosity >= 1:
            print 'No DB tests detected. Skipping Test DB creation...'
        self._db_patch = patch('django.db.backends.utils.CursorWrapper')
        self._db_mock = self._db_patch.start()
        self._db_mock.side_effect = RuntimeError('No testing the database!')
        return None

    def teardown_databases(self, *args, **kwargs):
        """
        Remove cursor patch.
        """
        if self._needs_db:
            return super(NoDatabaseMixin, self).teardown_databases(*args, **kwargs)
        self._db_patch.stop()
        return None


class FastTestRunner(NoDatabaseMixin, BaseRunner):
    """Actual test runner sub-class to make use of the mixin."""
